"""
Series of helper function to deal with healpix maps
"""
from __future__ import print_function, division


import numpy as np
import healpy as hp


# try: # prama: no cover
#     from wcsaxes import WCS # (deprecated)
# except ImportError: # pragma: no cover
from astropy.wcs import WCS

from astropy.io import fits
from astropy.wcs import utils as wcs_utils
from astropy.coordinates import ICRS, Galactic, SkyCoord, UnitSphericalRepresentation, Angle
from astropy import units as u

import logging
logging.basicConfig(format='%(asctime)s -- %(levelname)s: %(message)s', level=logging.DEBUG)

__all__ = ['build_WCS', 'build_WCS_lonlat',
           'build_WCS_cube', 'build_WCS_cube_lonlat',
           'build_WCS_2pts', 'build_ctype',
           'hp_is_nest', 'hp_celestial',
           'hp_to_wcs', 'hp_to_wcs_ipx',
           'hp_project',
           'gen_hpmap', 'build_hpmap',
           'equiv_celestial']

DEFAULT_shape_out = (512, 512)

VALID_PROJ = ['AZP', 'SZP', 'TAN', 'STG', 'SIN',
              'ARC', 'ZPN', 'ZEA', 'AIR', 'CYP',
              'CEA', 'CAR', 'MER', 'COP', 'COE',
              'COD', 'COO', 'SFL', 'PAR', 'MOL',
              'AIT', 'BON', 'PCO', 'TSC', 'CSC',
              'QSC','HPX','XPH']

VALID_GALACTIC = ['galactic', 'g']
VALID_EQUATORIAL = ['celestial2000','equatorial', 'eq', 'c', 'q', 'fk5']

def equiv_celestial(frame):
    """Return an equivalent ~astropy.coordfinates.builtin_frames

    Notes
    -----
    We do not care of the differences between ICRS/FK4/FK5
    """

    frame = frame.lower()
    if frame in VALID_GALACTIC:
        frame = Galactic()
    elif frame in VALID_EQUATORIAL:
        frame = ICRS()
    elif frame in [ 'ecliptic', 'e']:
        raise ValueError("Ecliptic coordinate frame not yet supported by astropy")
    return frame

def hp_celestial(hp_header):
    """Retrieve the celestial system used in healpix maps. From Healpix documentation this can have 3 forms :

    - 'EQ', 'C' or 'Q' : Celestial2000 = eQuatorial,
    - 'G' : Galactic
    - 'E' : Ecliptic,

    only Celestial and Galactic are supported right now as the Ecliptic coordinate system was just recently pulled to astropy

    Similar to :class:`~astropty.wcs.utils.wcs_to_celestial_frame` but for header from healpix maps

    Parameters
    ----------
    hp_header : :class:`~astropy.io.fits.header.Header`
        the header of the healpix map

    Returns
    -------
    frame : :class:`~astropy.coordinates.baseframe.BaseCoordinateFrame` subclass instance
        An instance of a :class:`~astropy.coordinates.baseframe.BaseCoordinateFrame`
        subclass instance that best matches the specified WCS.

    """
    coordsys = hp_header.get('COORDSYS')

    if coordsys:
        return equiv_celestial(coordsys)
    else:
        raise ValueError("No COORDSYS in header")

def hp_is_nest(hp_header):
    """Return True if the healpix header is in nested

    Parameters
    ----------
    hp_header : :class:`~astropy.io.fits.header.Header`
        the header of the healpix map

    Returns
    -------
    bool :
        True if the header is nested
    """

    ordering = hp_header.get('ORDERING')

    if ordering:
        if ordering.lower() == 'nested' or ordering.lower() == 'nest':
            return True
        elif ordering.lower() == 'ring':
            return False
        else:
            raise ValueError("Unknown ordering in healpix header")
    else:
        raise ValueError("Np ordering in healpix header")

def build_ctype(coordsys, proj_type):
    """Build a valid spatial ctype for a wcs header

    Parameters
    ----------
    coordsys : str ('GALATIC', 'EQUATORIAL')
        the coordinate system of the plate
    proj_type: str ('TAN', 'SIN', 'GSL', ...)
        any projection system supported by WCS

    Returns
    -------
    list:
        a list with the 2 corresponding spatial ctype
    """

    coordsys = coordsys.lower()
    proj_type = proj_type.upper()

    if not proj_type in VALID_PROJ:
        raise ValueError('Unvupported projection')

    if coordsys in VALID_GALACTIC:
        return [ coord+proj_type for coord in ['GLON-', 'GLAT-']]
    elif coordsys in VALID_EQUATORIAL:
        return [ coord+proj_type for coord in ['RA---', 'DEC--']]
    else:
        raise ValueError('Unsupported coordsys')

def build_WCS(coord, pixsize=0.01, shape_out=DEFAULT_shape_out, npix=None, proj_sys='EQUATORIAL', proj_type='TAN'):
    """Construct a :class:`~astropy.wcs.WCS` object for a 2D image

    Parameters
    ----------
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    pixsize : float
        size of the pixel (in degree)
    shape_out : tuple
        shape of the output map  (n_y,n_x)
    npix : int
        number of pixels in the final square map, the reference pixel will be at the center, superseed shape_out
    proj_sys : str ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the plate (from HEALPIX maps....)
    proj_type : str ('TAN', 'SIN', 'GSL', ...)
        the projection system to use

    Returns
    -------
    WCS: :class:`~astropy.wcs.WCS`
        An corresponding wcs object
    """

    proj_sys = proj_sys.lower()

    if proj_sys in VALID_EQUATORIAL:
        coord = coord.transform_to(ICRS)
        lon = coord.ra.deg
        lat = coord.dec.deg
    elif proj_sys in VALID_GALACTIC:
        coord = coord.transform_to(Galactic)
        lon = coord.l.deg
        lat = coord.b.deg
    else:
        raise ValueError('Unsuported coordinate system for the projection')

    proj_type = proj_type.upper()
    if not proj_type in VALID_PROJ:
        raise ValueError('Unvupported projection')

    if npix:
        shape_out = (npix, npix)

    w = WCS(naxis=2)

    # CRPIX IS in Fortran convention
    w.wcs.crpix = np.array(shape_out, dtype=np.float)/2
    w.wcs.cdelt = np.array([-pixsize, pixsize])
    w.wcs.crval = [lon, lat]

    w.wcs.ctype = build_ctype(proj_sys, proj_type)

    return w

def build_WCS_cube(coord, index, pixsize=0.01, shape_out=DEFAULT_shape_out, npix=None, proj_sys='EQUATORIAL', proj_type='TAN'):
    """Construct a :class:`~astropy.wcs.WCS` object for a 3D cube, where the 3rd dimension is an index

    Parameters
    ----------
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    index : int
        reference index
    pixsize : float
        size of the pixel (in degree)
    shape_out : tuple
        shape of the output map (n_y, n_x)
    npix : int
        number of pixels in the final map, the reference pixel will be at the center, override shape_out
    proj_sys : str ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the plate (from HEALPIX maps....)
    proj_type : str ('TAN', 'SIN', 'GSL', ...)
        the projection system to use

    Returns
    -------
    WCS: :class:`~astropy.wcs.WCS`
        An corresponding wcs object
    """

    proj_sys = proj_sys.lower()

    if proj_sys in VALID_EQUATORIAL:
        coord = coord.transform_to(ICRS)
        lon = coord.ra.deg
        lat = coord.dec.deg
    elif proj_sys in VALID_GALACTIC:
        coord = coord.transform_to(Galactic)
        lon = coord.l.deg
        lat = coord.b.deg
    else:
        raise ValueError('Unsuported coordinate system for the projection')

    proj_type = proj_type.upper()
    if not proj_type in VALID_PROJ:
        raise ValueError('Unvupported projection')


    if npix:
        shape_out = (npix, npix)

    w = WCS(naxis=3)
    w.wcs.crpix = np.append(np.array(shape_out, dtype=np.float)/2, 1)
    w.wcs.cdelt = np.append(np.array([-pixsize, pixsize]), 1)
    w.wcs.crval = np.array([lon, lat, index])

    w.wcs.ctype = np.append( build_ctype(proj_sys, proj_type), 'INDEX').tolist()

    return w


def _lonlat(build_WCS_func):
    """Will decorate the build_WCS function to be able to use it with lon/lat, proj_sys instead of an `astropy.coordinate.SkyCoord` object

    Parameters
    ----------
    build_WCS_func : fct
    a build_WCS_func function (with coords as the first argument

    Return
    ------
    The same function decorated

    Notes
    -----
    To use this decorator
    build_WCS_lonlat = _lonlat(build_WCS)
    or use @_lonlat on the function declaration

    """

    def decorator(lon, lat, src_frame='EQUATORIAL', **kargs):
        frame = equiv_celestial(src_frame)
        coord = SkyCoord(lon,lat, frame=frame, unit="deg")
        return build_WCS_func(coord, **kargs)

    decorator._coord = build_WCS_func
    decorator.__doc__ = str(build_WCS_func.__doc__).split('\n')[0] +\
    """
    Parameters
    ----------
    lon,lat : floats
        the sky coordinates of the center of projection
    src_frame :  str, ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the longitude and latitude
    """ + '\n'.join(str(build_WCS.__doc__).split('\n')[6:]) + \
    """
    Notes
    -----
    You can access a function using only catalogs with the ._coord() method
    """
    return decorator

build_WCS_lonlat = _lonlat(build_WCS)
build_WCS_cube_lonlat = _lonlat(build_WCS_cube)

def build_WCS_2pts(coords, pixsize=None, shape_out=DEFAULT_shape_out, npix=None, proj_sys='EQUATORIAL', proj_type='TAN', relative_pos=(2./5, 3./5)):
    """Construct a :class:`~astropy.wcs.WCS` object for a 2D image

    Parameters
    ----------
    coord : :class:`astropy.coordinate.SkyCoord`
        the 2 sky coordinates of the projection, they will be horizontal in the resulting wcs
    pixsize : float
        size of the pixel (in degree) (default: None, use relative_pos and shape_out)
    shape_out : tuple
        shape of the output map  (n_y,n_x)
    npix : int
        number of pixels in the final square map, superseed shape_out
    coordsys : str ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the plate (from HEALPIX maps....) will be rotated anyway
    proj_type : str ('TAN', 'SIN', 'GSL', ...)
        the projection system to use, the first coordinate will be the projection center
    relative_pos : tuple
        the relative position of the 2 sources along the x direction [0-1] (will be computed if pixsize is given)

    Returns
    -------
    WCS: :class:`~astropy.wcs.WCS`
        An corresponding wcs object

    Notes
    -----

    By default relative_pos is used to place the sources, and the
    pixsize is derived, but if you define pixsize, then the
    relative_pos will be computed and the sources placed at the center
    of the image

    """

    if npix:
        shape_out = (npix, npix)

    w = WCS(naxis=2)

    if not pixsize:
        # Compute pixsize from relative_pos and distance between
        # sources
        pix_distance = np.array(relative_pos)*shape_out[1]
        pix_distance = pix_distance.max() - pix_distance.min()
        ang_distance = coords[0].separation(coords[1])
        pixsize = ang_distance.deg / pix_distance
    else:
        # Compute relative_pos from pixsize and distance between
        # sources
        ang_distance = coords[0].separation(coords[1])
        pix_distance = ang_distance.deg / pixsize
        relative_pos = pix_distance / shape_out[1]
        # Center it
        relative_pos = 0.5+np.array([-1.,1])*relative_pos/2

    proj_sys = proj_sys.lower()
    if proj_sys in VALID_EQUATORIAL:
        coords = [ coord.transform_to(ICRS) for coord in coords]
        lon = coords[0].ra.deg
        lat = coords[0].dec.deg
    elif proj_sys in VALID_GALACTIC:
        coords = [ coord.transform_to(Galactic) for coord in coords]
        lon = coords[0].l.deg
        lat = coords[0].b.deg
    else:
        raise ValueError('Unsuported coordinate system for the projection')

    proj_type = proj_type.upper()
    if not proj_type in VALID_PROJ:
        raise ValueError('Unvupported projection')

    # Put the first source on the relative_pos[0]
    w.wcs.crpix = np.array([relative_pos[0],1./2], dtype=np.float)*(np.array(shape_out, dtype=np.float)[::-1])
    w.wcs.crval = [lon, lat]

    w.wcs.cdelt = np.array([-pixsize, pixsize])
    # Computes the on-sky position angle (East of North) between this SkyCoord and another.
    rot_angle = (coords[0].position_angle(coords[1])+Angle(90, unit='deg')).wrap_at('180d')

    logging.debug('... rotating frame with %s deg'%rot_angle.degree)
    w.wcs.pc = [ [np.cos(rot_angle.radian) , np.sin(-rot_angle.radian)],
                 [np.sin(rot_angle.radian) , np.cos(rot_angle.radian)] ]

    w.wcs.ctype = build_ctype(proj_sys, proj_type)

    return w

def hp_to_wcs(hp_map, hp_header, w, shape_out=DEFAULT_shape_out, npix=None, order=0):
    """Project an Healpix map on a wcs header, using nearest neighbors.

    Parameters
    ----------
    hp_map : array_like
        healpix map with
    hp_header : :class:`astropy.fits.header.Header`
        corresponding header to project on
    w : :class:`astropy.wcs.WCS`
        wcs object to project with
    shape_out : tuple
        shape of the output map (n_y, n_x)
    npix : int
        number of pixels in the final square map, superseed shape_out
    order : int (0|1)
        order of the interpolation 0: nearest-neighbor, 1: bi-linear interpolation

    Return
    ------
    array_like
        the projected map in a 2D array of shape shape_out
    """

    if npix:
        shape_out = (npix, npix)

    # Array of pixels centers -- from the astropy.wcs documentation :
    #
    # Here, *origin* is the coordinate in the upper left corner of the
    # image.In FITS and Fortran standards, this is 1.  In Numpy and C
    # standards this is 0.
    #
    # This as we are using the C convention, the center of the first pixel is -0.5, -0.5
    yy, xx = np.indices(shape_out)-0.5


    alon, alat = w.wcs_pix2world(xx,yy,0)
    mask = ~np.logical_or(np.isnan(alon),np.isnan(alat)) # if pixel lies outside of projected area
    proj_map = np.ma.array(np.zeros(shape_out), mask=~mask, fill_value=np.nan)

    # Determine if we need a rotation for different coordinate systems
    frame_in =  hp_celestial(hp_header)
    frame_out = wcs_utils.wcs_to_celestial_frame(w)

    if not frame_in.is_equivalent_frame(frame_out):
        logging.debug('... converting coordinate system')
        coords = frame_out.realize_frame(UnitSphericalRepresentation(alon*u.deg, alat*u.deg)).transform_to(frame_in)
        alon = coords.data.lon.deg
        alat = coords.data.lat.deg

    # Healpix conventions...
    phi = np.radians(alon)
    theta = np.radians(90-alat)


    if order == 0:
        ipix = hp.ang2pix(hp.npix2nside(len(hp_map)), theta[mask], phi[mask], nest=hp_is_nest(hp_header))
        proj_map[mask] = hp_map[ipix]
    elif order == 1:
        proj_map[mask] = hp.get_interp_val(hp_map, theta[mask], phi[mask], nest=hp_is_nest(hp_header))
    else:
        raise ValueError("Unsupported order for the interpolation")

    return proj_map.filled()

def hp_to_wcs_ipx(hp_header, w, shape_out=DEFAULT_shape_out, npix=None):
    """Return the indexes of pixels of a given wcs and shape_out,
    within a nside healpix map.

    Parameters
    ----------
    hp_header : :class:`astropy.fits.header.Header`
        header of the healpix map, should contain nside and coordsys and ordering
    w : :class:`astropy.wcs.WCS`
        wcs object to project with
    shape_out : tuple
        shape of the output map (n_y, n_x)
    npix : int
        number of pixels in the final square map, superseed shape_out

    Return
    ------
    2D array_like
        mask for the given map
    array_like
        corresponding pixel indexes

    Notes
    -----
    The map could then easily be constructed using

    proj_map = np.ma.array(np.zeros(shape_out), mask=~mask, fill_value=np.nan)
    proj_map[mask] = healpix_map[ipix]
    """

    if npix:
        shape_out = (npix, npix)

    # Array of pixels centers -- from the astropy.wcs documentation :
    #
    # Here, *origin* is the coordinate in the upper left corner of the
    # image.In FITS and Fortran standards, this is 1.  In Numpy and C
    # standards this is 0.
    #
    # This as we are using the C convention, the center of the first pixel is -0.5, -0.5
    yy, xx = np.indices(shape_out)-0.5

    alon, alat = w.wcs_pix2world(xx,yy,0)
    mask = ~np.logical_or(np.isnan(alon),np.isnan(alat)) # if pixel lies outside of projected area

    frame_in =  hp_celestial(hp_header)
    frame_out = wcs_utils.wcs_to_celestial_frame(w)

    nside = hp_header['NSIDE']

    if not frame_in.is_equivalent_frame(frame_out):
        logging.debug('... converting coordinate system')
        coords = frame_out.realize_frame(UnitSphericalRepresentation(alon*u.deg, alat*u.deg)).transform_to(frame_in)
        alon = coords.data.lon.deg
        alat = coords.data.lat.deg

    # Healpix conventions...
    phi = np.radians(alon)
    theta = np.radians(90-alat)

    ipix = hp.ang2pix(nside, theta[mask], phi[mask], nest=hp_is_nest(hp_header))

    return mask, ipix

def hp_project(hp_map, hp_header, coord, pixsize=0.01, npix=512, proj_sys='GALACTIC', proj_type='TAN', order=0, hdu=False):
    """Project an healpix map at a single given position

    Parameters
    ----------
    hp_map : array_like
        healpix map with
    hp_header : :class:`astropy.fits.header.Header`
        corresponding header to project on
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    pixsize : float
        size of the pixel (in degree)
    npix : int
        number of pixels in the final map, the reference pixel will be at the center
    proj_sys : str, ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the projection (from HEALPIX maps....)
    proj_type : str ('TAN', 'SIN', 'GSL', ...)
        the projection system to use
    order : int (0|1)
        order of the interpolation 0: nearest-neighbor, 1: bi-linear interpolation
    hdu : bool
        return a :class:`astropy.io.fits.PrimaryHDU` instead of just a ndarray

    Return
    ------
    array_like
        2D images at lon, lat position
    or :class:`astropy.io.fits.PrimaryHDU` (optionnal)
        containing the array and the corresponding header
    """

    w = build_WCS(coord, pixsize, npix=npix, proj_sys=proj_sys, proj_type=proj_type)
    proj_map = hp_to_wcs(hp_map, hp_header, w, npix=npix, order=order)

    if hdu:
        return fits.PrimaryHDU(proj_map, w.to_header(relax=0x20000))
    else:
        return proj_map

def gen_hpmap(maps):
    """Generator function for large maps and low memory system

    Parameters
    ----------
    maps : list
        A list of Nmap tuples with either:
            * (filename, path_to_localfilename, healpix header)
            * (filename, healpix vector, healpix header)

    Return
    ------
    tuple
        Return a tuple (filename, healpix map, healpix header) corresponding to the inputed list
    """
    for filename, iMap, iHeader in maps:
        if isinstance(iMap, str):
            iMap = hp.read_map(iMap, verbose=False)
            iMap = hp.ma(iMap)
        yield (filename, iMap, iHeader)

def build_hpmap(filenames, low_mem=True):
    """From a filename list, build a tuple usable with gen_hmap()

    Parameters
    ----------
    filenames: list
        A list of Nmap filenames of healpix maps
    low_mem : bool
        On low memory system, do not read the maps themselves (default: only header)

    Return
    ------
    tuple list
        A list of tuple which can be used by gen_hpmap    """

    hp_maps = []
    for filename in filenames:
        hp_header = fits.getheader(filename, 1)
        if low_mem == True:
            hp_map = filename
        else:
            hp_map = hp.read_map(filename, verbose=False)
            hp_map = hp.ma(hp_map)
        hp_maps.append( (filename, hp_map, hp_header ) )
    return hp_maps

def group_hpmap(maps):
    """Group into a dictionnary the hp_maps according to map properties

    Parameters
    ----------
    hpmap: tuple list
        A list of tuple which can be used by gen_hpmap

    Return
    ------
    dict
        A dictionnary where each key contains hpmap sharing the same properties

    """
    grouped_hpmaps = {}

    for (filename, iMap, iHeader) in maps:
        mapKey = "%s_%s_%s"%(iHeader['NSIDE'], iHeader['ORDERING'],  hp_celestial(iHeader).name)
        iHeader['mapKey'] = mapKey

        if mapKey in grouped_hpmaps.keys():
            grouped_hpmaps[mapKey].append((filename, iMap, iHeader))
        else:
            grouped_hpmaps[mapKey] = [(filename, iMap, iHeader)]

    return grouped_hpmaps



#! /usr/bin/env python

import logging
logger = logging.getLogger('django')

import warnings
import os, sys
import argparse

import numpy as np
import healpy as hp

# try: # prama: no cover
#     from wcsaxes import WCS # (deprecated)
# except ImportError: # pragma: no cover
from astropy.wcs import WCS

from astropy.io import fits
from astropy import units as u
from astropy.coordinates import SkyCoord
from photutils import CircularAperture
from photutils import aperture_photometry

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

try: # pragma: py3
    from configparser import ConfigParser
except ImportError: # pragma: py2
    from ConfigParser import ConfigParser

try: # pragma: py3
    from io import BytesIO
except ImportError: # pragma: py2
    from cStringIO import StringIO as BytesIO

from base64 import b64encode

try: # pragma: py3
    FileNotFoundError
except NameError: # pragma: py2
    FileNotFoundError = IOError

from .hp_helper import build_WCS, hp_to_wcs_ipx, VALID_PROJ
from .hp_helper import equiv_celestial
from .hp_helper import build_hpmap, group_hpmap, gen_hpmap

DEFAULT_npix = 256
DEFAULT_pixsize = 1
DEFAULT_coordframe = 'galactic'
DEFAULT_ctype = 'TAN'

class CutSkySquare:
    """
    Container for Healpix maps and cutsky methods

    ...

    Attributes
    ----------
    npix : int
        the number of pixels for the square maps
    pixsize : float
        the size of the pixels [arcmin]
    ctype : str
        a valid projection type (default : TAN)
    maps : dictonnary
        a grouped dictionnary of gen_hpmap tuples (filename, map, header)


    """
    def __init__(self, maps, npix=DEFAULT_npix, pixsize=DEFAULT_pixsize, ctype=DEFAULT_ctype, low_mem=True):
        """Initialization of a CutSkySquare class

        Parameters
        ----------
        maps : list of tuple
            list of tuple (filename, {opt}) where filename is the full
            path to the healpix map and {opt} is a dictionnary with
            "optional" header for the file (see Notes).
        npix : int
            the number of pixels for the square maps
        pixsize : float
            the size of the pixels [arcmin]
        ctype : str
            a valid projection type (default : TAN)
        low_mem: bool
            a boolean deciding if all the maps are loaded in memory or
            not (default: True, the maps are not loaded on init)

        Notes
        -----

        The {opt} dictionnary MUST containt, at least, the key
        'legend' which will be used to uniquely identify the cutted
        map, other possible keys 'doContour' with a boolean value to
        contour the map latter one

        """

        # Define the basic parameters for the output maps
        self.npix = npix
        self.pixsize = pixsize
        self.ctype = ctype

        # Build an hp_map list and extend the headers of the maps with
        # the optionnal values
        filenames, opts = [filename for filename, opt in maps], [opt for filename, opt in maps]
        hp_map = build_hpmap(filenames, low_mem=low_mem)
        for (filename, iMap, iHeader), opt in zip(hp_map, opts):
            iHeader.extend(opt)

        # group them by map properties for efficiencies reasons
        self.maps = group_hpmap(hp_map)

        # Save intermediate results
        self.cut_fits = None
        self.lonlat = [0,0]
        self.coordframe = DEFAULT_coordframe

    def cutsky_fits(self,lonlat = [0, 0], coordframe = DEFAULT_coordframe):
        """Efficiently cut the healpix maps and return cutted fits file with proper header

        Parameters
        ----------
        lonlat : array of 2 floats
            the longitude and latitude of the center of projection [deg]
        coordframe : str
            the coordinate frame used for the position AND the projection

        Returns
        -------
        list of dictionnaries
            the dictionnary as two keys 'legend' (the opts{'legend'} see
            __init()) and 'fits' an ~astropy.io.fits.ImageHDU
        """

        # Center of projection
        coord_in = SkyCoord(lonlat[0],lonlat[1], unit=u.deg, frame=equiv_celestial(coordframe))

        # Build the target WCS header
        w = build_WCS(coord_in, pixsize=self.pixsize/60., npix=self.npix, proj_sys=coordframe, proj_type=self.ctype)

        cut_fits = []
        for group in self.maps.keys():
            logger.info('projecting '+group)

            maps = self.maps[group]

            # Construct a basic healpix header from the group key, this
            # will be commun for all the maps in this group
            hp_header = maps[0][2]

            # Extract mask & pixels, common for all healpix maps of this
            # group
            mask, ipix = hp_to_wcs_ipx(hp_header, w, npix=self.npix)

            # Set up the figure, common for all healpix maps of this group
            logger.info('cutting maps')

            # Now the actual healpix map reading and projection
            for filename, iMap, iHeader in gen_hpmap(maps):
                legend = iHeader['legend']
                patch = np.ma.array(np.zeros((self.npix, self.npix)), mask=~mask, fill_value=np.nan)
                patch[mask] = iMap[ipix]
                header = w.to_header()
                header.append(('filename',filename) ) #, 'original healpix filename')
                header.append(('legend', legend) ) #, 'cutsky legend')
                if 'doContour' in iHeader.keys():
                    header.append(('doContour', iHeader['doContour']))

                cut_fits.append( {'legend': legend,
                                  'fits': fits.ImageHDU(patch, header)} )
        self.lonlat = lonlat
        self.coordframe = coordframe
        self.cut_fits = cut_fits

        return cut_fits

    def cutsky_png(self, lonlat = [0, 0], coordframe = DEFAULT_coordframe):
        """Efficiently cut the healpix maps and return cutted fits file with proper header and corresponding png

        Parameters
        ----------
        lonlat : array of 2 floats
            the longitude and latitude of the center of projection [deg]
        coordframe : str
            the coordinate frame used for the position AND the projection

        Returns
        -------
        list of dictionnaries
            the dictionnary as two keys 'legend' (the opts{'legend'} see
            __init()), 'fits' an ~astropy.io.fits.ImageHDU,
            'png', a b61encoded png image of the fits

        """

        if self.lonlat == lonlat and \
           self.coordframe == coordframe and  \
           self.cut_fits:
            # Retrieve previously cut maps
            cut_fits = self.cut_fits
        else:
            # Or cut the maps
            cut_fits = self.cutsky_fits(lonlat=lonlat, coordframe=coordframe)

        # Common WCS for all cut maps
        w = WCS(cut_fits[0]['fits'])

        # Plotting
        patch = np.zeros((self.npix, self.npix))
        fig=plt.figure()

        wcs_proj=WCS(w.to_header())
        ax_wcs=fig.add_axes([0.1,0.1,0.9,0.9],projection=wcs_proj)
        proj_im = ax_wcs.imshow(patch, interpolation='none', origin='lower' )
        ax_wcs.coords.grid(color='green', linestyle='solid', alpha=0.5)

        if np.str(coordframe)=="EQUATORIAL":
            ax_wcs.coords['ra'].set_ticks(color='white')
            ax_wcs.coords['dec'].set_ticks(color='white')
            ax_wcs.coords['ra'].set_axislabel(r'$\alpha_\mathrm{J2000}$')
            ax_wcs.coords['dec'].set_axislabel(r'$\delta_\mathrm{J2000}$')
        elif np.str(coordframe)=="GALACTIC":
            ax_wcs.coords['glon'].set_ticks(color='red')
            ax_wcs.coords['glat'].set_ticks(color='red')
            ax_wcs.coords['glon'].set_axislabel(r'$l$')
            ax_wcs.coords['glat'].set_axislabel(r'$b$')

        for cut_fit in cut_fits :

            legend = cut_fit['legend']
            logger.debug('plotting '+ legend)

            patch = cut_fit['fits'].data
            patch_header = cut_fit['fits'].header

            proj_im.set_data(patch)
            proj_im.set_clim(vmin=patch.min(), vmax=patch.max())

            if 'doContour' in patch_header.keys() and patch_header['doContour']:
                logger.debug('contouring '+legend)
                levels=[patch.max()/3., patch.max()/2.]
                if ((patch.max()-patch.mean()) > 3*patch.std()):
                    proj_cont = ax_wcs.contour(patch,levels=levels,colors="white",interpolation='bicubic')
                else:
                    proj_cont = None

            logger.debug('saving '+legend)

            # Add the map to the cut_fit dictionnary
            output_map = BytesIO()
            plt.savefig(output_map,bbox_inches='tight', format='png',dpi=75, frameon=False)
            cut_fit['png'] = b64encode(output_map.getvalue()).strip()
            logger.debug(legend+ ' done')

        return cut_fits

    def cutsky_phot(self, lonlat = [0, 0], coordframe = DEFAULT_coordframe):
        """Efficiently cut the healpix maps and return cutted fits file with proper header and corresponding photometry

        Parameters
        ----------
        lonlat : array of 2 floats
            the longitude and latitude of the center of projection [deg]
        coordframe : str
            the coordinate frame used for the position AND the projection

        Returns
        -------
        list of dictionnaries
            the dictionnary as two keys 'legend' (the opts{'legend'} see
            __init()), 'fits' an ~astropy.io.fits.ImageHDU,
            'phot', the corresponding photometry

        """

        if self.lonlat == lonlat and \
           self.coordframe == coordframe and  \
           self.cut_fits:
            # Retrieve previously cut maps
            cut_fits = self.cut_fits
        else:
            # Or cut the maps
            cut_fits = self.cutsky_fits(lonlat=lonlat, coordframe=coordframe)

        positions = [(self.npix*1./2., self.npix*1./2) ]
        apertures = CircularAperture(positions, r = 3./self.pixsize)

        for cut_fit in cut_fits :
            legend = cut_fit['legend']
            logger.debug('phot on '+ legend)

            patch = cut_fit['fits'].data
            cut_fit['phot'] = aperture_photometry(patch-np.median(patch), apertures)

            logger.debug(legend+ ' done')

        return cut_fits


def cutsky(lonlat=[0, 0], patch=[256, 1], coordframe='galactic', ctype=DEFAULT_ctype, maps=None):
    """Old interface to cutsky -- Here for compability"""

    warnings.warn("deprecated", DeprecationWarning)

    if not maps:
        raise FileNotFoundError("No healpix map to project")

    # Transform the way we defined maps
    new_maps = []
    for key in maps.iterkeys():
        filename = maps[key]['filename']
        opt = {'legend': key}
        if maps[key].has_key('doContour'):
            opt['doContour'] = maps[key]['doContour']
        new_maps.append((filename, opt))

    CutThoseMaps = CutSkySquare(new_maps, npix=patch[0], pixsize=patch[1], ctype=ctype)
    result = CutThoseMaps.cutsky_png(lonlat=lonlat, coordframe=coordframe)
    result = CutThoseMaps.cutsky_phot(lonlat=lonlat, coordframe=coordframe)

    return result

def parse_args(args):
    """Parse arguments from the command line"""
    parser = argparse.ArgumentParser(description="Reproject the spherical sky onto a plane.")
    parser.add_argument('lon', type=float,
                        help='longitude of the projection [deg]')
    parser.add_argument('lat', type=float,
                        help='latitude of the projection [deg]')

    # Removed the default argument from here otherwise the config file
    # wont be used

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--npix', type=int,
                        help='number of pixels (default 256)')
    group.add_argument('--radius', type=float,
                       help='radius of the requested region [deg] ')

    parser.add_argument('--pixsize', type=float,
                        help='pixel size [arcmin] (default 1)')

    parser.add_argument('--coordframe', required=False,
                        help='coordinate frame of the lon. and \
                        lat. of the projection and the projected map \
                        (default: galactic)',
                        choices=['galactic', 'fk5'])
    parser.add_argument('--ctype', required=False,
                        help='any projection code supported by wcslib\
                         (default:TAN)',
                        choices=VALID_PROJ)

    parser.add_argument('--mapfilenames', nargs='+', required=False,
                        help='Absolute path to the healpix maps')

    verb = parser.add_mutually_exclusive_group()
    verb.add_argument('-v','--verbose', action='store_true', help='verbose mode')
    verb.add_argument('-q','--quiet', action='store_true', help='quiet mode')

    parser.add_argument('--fits', action='store_true', help='output fits file')
    parser.add_argument('--png', action='store_true', help='output png file')
    parser.add_argument('--votable', action='store_true', help='output votable file')

    parser.add_argument('--conf', required=False,
                        help='Absolute path to a config file')



    # Do the actual parsing
    parsed_args = parser.parse_args(args)
    #parsed_args = parser.parse_args('0 0'.split())

    # Put the list of filenames into the same structure as the config
    # file, we are loosing the doContour keyword but...
    if parsed_args.mapfilenames:
        parsed_args.maps = [ ( filename,
                               dict( [('legend', os.path.basename(filename)) ]) )
                             for filename in
                             parsed_args.mapfilenames ]
    else:
        parsed_args.maps = None

    if parsed_args.verbose:
        parsed_args.verbosity = logging.DEBUG
    elif parsed_args.quiet:
        parsed_args.verbosity = logging.ERROR
    else:
        parsed_args.verbosity = None

    return parsed_args


def parse_config(conffile=None):
    """Parse options from a configuration file."""
    config = ConfigParser()

    # Look for cutsky.cfg at several locations
    conffiles = [ os.path.join(directory, 'cutsky.cfg') for directory
                  in [os.curdir, os.path.join(os.path.expanduser("~"), '.config/cutsky')] ]

    # If specifically ask for a config file, then put it at the
    # beginning ...
    if conffile:
        conffiles.insert(0, conffile)

    found = config.read(conffiles)

    if not found:
        raise FileNotFoundError

    # Basic options, same as the option on the command line
    options = {}
    if config.has_option('cutsky','npix'):
        options['npix'] = config.getint('cutsky', 'npix')
    if config.has_option('cutsky', 'pixsize'):
        options['pixsize'] = config.getfloat('cutsky', 'pixsize')
    if config.has_option('cutsky', 'coordframe'):
        options['coordframe'] = config.get('cutsky', 'coordframe')
    if config.has_option('cutsky', 'ctype'):
        options['ctype'] = config.get('cutsky', 'ctype')
    if config.has_option('cutsky', 'verbosity'):
        level = config.get('cutsky', 'verbosity').lower()
        allowed_levels = { 'verbose': logging.DEBUG,
                           'debug': logging.DEBUG,
                           'quiet': logging.ERROR,
                           'error': logging.ERROR,
                           'info': logging.INFO}
        if level in allowed_levels.keys():
            options['verbosity'] = allowed_levels[level]
        else:
            try:
                options['verbosity'] = int(level)
            except ValueError:
                options['verbosity'] = None


    if config.has_option('cutsky', 'fits') and config.get('cutsky', 'fits'):
        options['fits'] = True

    if config.has_option('cutsky', 'png') and config.get('cutsky', 'png'):
        options['png'] = True

    if config.has_option('cutsky', 'votable') and config.get('cutsky', 'votable'):
        options['votable'] = True

    # Map list, only get the one which will be projected
    # Also check if contours are requested
    mapsToCut = []
    for section in config.sections():
        if section != 'cutsky':
            if config.has_option(section, 'doCut') and config.has_option(section, 'filename'):
                if config.getboolean(section, 'doCut'):
                    filename = config.get(section, 'filename')
                    opt = {'legend': section}
                    if config.has_option(section, 'doContour'):
                        opt['doContour'] = config.getboolean(section, 'doContour')
                    mapsToCut.append((filename, opt))

    options['maps'] = mapsToCut

    return options

def combine_args(args, config):
    """
    Combine the different sources of arguments (command line,
    configfile or default arguments
    """

    # This is where the default arguments are set
    pixsize = args.pixsize or config.get('pixsize') or \
              DEFAULT_pixsize
    coordframe = args.coordframe or config.get('coordframe') or \
                 DEFAULT_coordframe
    ctype = args.ctype or config.get('ctype') or \
                 DEFAULT_ctype

    verbosity = args.verbosity or config.get('verbosity') or logging.INFO
    logger.setLevel(verbosity)

    # npix and radius are mutually exclusive, thus if radius is set we
    # need to compute npix
    if args.radius:
        args.npix = int(args.radius/(float(pixsize)/60))

    npix = args.npix or config.get('npix') or \
           DEFAULT_npix

    maps = args.maps or config.get('maps') or \
           None

    output = {}
    output['fits'] = args.fits or config.get('fits') or False
    output['png'] = args.png or config.get('png') or True
    output['votable'] = args.votable or config.get('votable') or False

    return npix, pixsize, coordframe, ctype, maps, output


def main():

    from base64 import b64decode

    try:
        args = parse_args(sys.argv[1:])
    except SystemExit:
        sys.exit()

    try:
        config = parse_config(args.conf)
    except FileNotFoundError:
        config = {}

    npix, pixsize, coordframe, ctype, maps, output = combine_args(args, config)

    CutThoseMaps = CutSkySquare(maps, npix=npix, pixsize=pixsize, ctype=ctype)
    if output['fits']:
        results = CutThoseMaps.cutsky_fits(lonlat=[args.lon, args.lat], coordframe=coordframe)
    if output['png']:
        results = CutThoseMaps.cutsky_png(lonlat=[args.lon, args.lat], coordframe=coordframe)
    if output['votable']:
        results = CutThoseMaps.cutsky_phot(lonlat=[args.lon, args.lat], coordframe=coordframe)

    for result in results:
        if 'fits' in result.keys() and output['fits']:
            try:
                hdulist = fits.HDUList([ fits.PrimaryHDU(), result['fits'] ])
                hdulist.writeto(result['legend']+'.fits', clobber=True)
            except NotImplementedError:
                result['fits'].data = result['fits'].data.filled()
                hdulist = fits.HDUList([ fits.PrimaryHDU(), result['fits'] ])
                hdulist.writeto(result['legend']+'.fits', clobber=True)

        if 'png' in result.keys() and output['png']:
            output = open(result['legend']+'.png', 'wb')
            output.write(b64decode(result['png']))
            output.close()

        if 'phot' in result.keys() and output['votable'] :
            result['phot'].write(result['legend']+'.xml', format='votable')

if __name__ == '__main__':
    main()



import os
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

opts = dict(name="hpproj",
            author='Alexandre Beelen',
            author_email='alexandre.beelen@ias.u-psud.fr',
            maintainer="Marian Douspis",
            maintainer_email="marian.douspis@ias.u-psud.fr",
            description='Projection of Healpix maps onto a planar grid',
            long_description='Projection of Healpix maps onto a planar grid using wcs headers',
            url='https://git.ias.u-psud.fr/abeelen/hpproj',
            download_url='https://git.ias.u-psud.fr/abeelen/hpproj/repository/archive.tar.gz?0.2.1',
            license='LGPL-3.0+',
            classifiers=['Topic :: Scientific/Engineering :: Astronomy',
                         'Intended Audience :: Science/Research',
                         'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)'],
            version='0.2.1',
            packages=['hpproj'],
            package_dir={'hpproj'  : 'hpproj'},
            entry_points = {
                'console_scripts': [
                    'cutsky = hpproj.cutsky:main'] },
            setup_requires=['pytest-runner'],
            tests_require=['pytest'],

            requires=['numpy', 'matplotlib', 'healpy', 'astropy',
                      'photutils' ],
)


if __name__ == '__main__':
    setup(**opts)
